# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from paginate_sqlalchemy import SqlalchemyOrmPage

from ziggurat_foundations.models.base import get_db_session
from ziggurat_foundations.models.services import BaseService
from ziggurat_foundations.permissions import (
    ANY_PERMISSION,
    ALL_PERMISSIONS,
    PermissionTuple,
)

__all__ = ["GroupService"]


class GroupService(BaseService):
    @classmethod
    def get(cls, group_id, db_session=None):
        """
        Fetch row using primary key -
        will use existing object in session if already present

        :param group_id:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        return db_session.query(cls.model).get(group_id)

    @classmethod
    def by_group_name(cls, group_name, db_session=None):
        """
        fetch group by name

        :param group_name:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model).filter(cls.model.group_name == group_name)
        return query.first()

    @classmethod
    def get_user_paginator(
        cls,
        instance,
        page=1,
        item_count=None,
        items_per_page=50,
        user_ids=None,
        GET_params=None,
    ):
        """
        returns paginator over users belonging to the group

        :param instance:
        :param page:
        :param item_count:
        :param items_per_page:
        :param user_ids:
        :param GET_params:
        :return:
        """
        if not GET_params:
            GET_params = {}
        GET_params.pop("page", None)
        query = instance.users_dynamic
        if user_ids:
            query = query.filter(cls.models_proxy.UserGroup.user_id.in_(user_ids))
        return SqlalchemyOrmPage(
            query,
            page=page,
            item_count=item_count,
            items_per_page=items_per_page,
            **GET_params
        )

    @classmethod
    def resources_with_possible_perms(
        cls,
        instance,
        perm_names=None,
        resource_ids=None,
        resource_types=None,
        db_session=None,
    ):
        """
        returns list of permissions and resources for this group,
            resource_ids restricts the search to specific resources

        :param instance:
        :param perm_names:
        :param resource_ids:
        :param resource_types:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session, instance)

        query = db_session.query(
            cls.models_proxy.GroupResourcePermission.perm_name,
            cls.models_proxy.Group,
            cls.models_proxy.Resource,
        )
        query = query.filter(
            cls.models_proxy.Resource.resource_id
            == cls.models_proxy.GroupResourcePermission.resource_id
        )
        query = query.filter(
            cls.models_proxy.Group.id
            == cls.models_proxy.GroupResourcePermission.group_id
        )
        if resource_ids:
            query = query.filter(
                cls.models_proxy.GroupResourcePermission.resource_id.in_(resource_ids)
            )

        if resource_types:
            query = query.filter(
                cls.models_proxy.Resource.resource_type.in_(resource_types)
            )

        if perm_names not in ([ANY_PERMISSION], ANY_PERMISSION) and perm_names:
            query = query.filter(
                cls.models_proxy.GroupResourcePermission.perm_name.in_(perm_names)
            )
        query = query.filter(
            cls.models_proxy.GroupResourcePermission.group_id == instance.id
        )

        perms = [
            PermissionTuple(
                None, row.perm_name, "group", instance, row.Resource, False, True
            )
            for row in query
        ]
        for resource in instance.resources:
            perms.append(
                PermissionTuple(
                    None, ALL_PERMISSIONS, "group", instance, resource, True, True
                )
            )
        return perms



"""increase user and token field sized

Revision ID: b5e6dd3449dd
Revises: 57bbf0c387c
Create Date: 2016-05-05 11:00:59.915215

"""
from __future__ import unicode_literals

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b5e6dd3449dd"
down_revision = "57bbf0c387c"


def upgrade():
    op.alter_column(
        "external_identities",
        "access_token",
        type_=sa.Unicode(512),
        existing_type=sa.Unicode(255),
    )
    op.alter_column(
        "external_identities",
        "alt_token",
        type_=sa.Unicode(512),
        existing_type=sa.Unicode(255),
    )
    op.alter_column(
        "external_identities",
        "token_secret",
        type_=sa.Unicode(512),
        existing_type=sa.Unicode(255),
    )
    op.alter_column(
        "users", "user_name", type_=sa.Unicode(128), existing_type=sa.Unicode(32)
    )
    op.alter_column(
        "groups_permissions",
        "perm_name",
        type_=sa.Unicode(64),
        existing_type=sa.Unicode(30),
        nullable=False,
    )
    op.alter_column(
        "groups_resources_permissions",
        "perm_name",
        type_=sa.Unicode(64),
        existing_type=sa.Unicode(50),
        nullable=False,
    )
    op.alter_column(
        "users_permissions",
        "perm_name",
        type_=sa.Unicode(64),
        existing_type=sa.Unicode(30),
        nullable=False,
    )
    op.alter_column(
        "users_resources_permissions",
        "perm_name",
        type_=sa.Unicode(64),
        existing_type=sa.Unicode(50),
        nullable=False,
    )


def downgrade():
    pass



"""initial table layout

Revision ID: 2bb1ba973f0b
Revises: None
Create Date: 2011-11-10 22:32:14.464939

"""
from __future__ import unicode_literals

import sqlalchemy as sa
from alembic import op

# downgrade revision identifier, used by Alembic.
revision = "2bb1ba973f0b"
down_revision = None


def upgrade():
    op.create_table(
        "groups",
        sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
        sa.Column("group_name", sa.Unicode(50), unique=True),
        sa.Column("description", sa.Text()),
        sa.Column("member_count", sa.Integer, nullable=False, default=0),
    )
    op.create_table(
        "groups_permissions",
        sa.Column(
            "group_name",
            sa.Unicode(50),
            sa.ForeignKey("groups.group_name", onupdate="CASCADE", ondelete="CASCADE"),
            primary_key=True,
        ),
        sa.Column("perm_name", sa.Unicode(30), primary_key=True),
    )

    op.create_table(
        "users",
        sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
        sa.Column("user_name", sa.Unicode(30), unique=True),
        sa.Column("user_password", sa.Unicode(40)),
        sa.Column("email", sa.Unicode(100), nullable=False, unique=True),
        sa.Column("status", sa.SmallInteger(), nullable=False),
        sa.Column("security_code", sa.Unicode(40), default="default"),
        sa.Column(
            "last_login_date",
            sa.TIMESTAMP(timezone=False),
            default=sa.sql.func.now(),
            server_default=sa.func.now(),
        ),
    )

    op.create_table(
        "users_permissions",
        sa.Column(
            "user_name",
            sa.Unicode(50),
            sa.ForeignKey("users.user_name", onupdate="CASCADE", ondelete="CASCADE"),
            primary_key=True,
        ),
        sa.Column("perm_name", sa.Unicode(30), primary_key=True),
    )

    op.create_table(
        "users_groups",
        sa.Column(
            "group_name",
            sa.Unicode(50),
            sa.ForeignKey("groups.group_name", onupdate="CASCADE", ondelete="CASCADE"),
            primary_key=True,
        ),
        sa.Column(
            "user_name",
            sa.Unicode(30),
            sa.ForeignKey("users.user_name", onupdate="CASCADE", ondelete="CASCADE"),
            primary_key=True,
        ),
    )

    op.create_table(
        "resources",
        sa.Column(
            "resource_id",
            sa.BigInteger(),
            primary_key=True,
            nullable=False,
            autoincrement=True,
        ),
        sa.Column("resource_name", sa.Unicode(100), nullable=False),
        sa.Column("resource_type", sa.Unicode(30), nullable=False),
        sa.Column(
            "owner_group_name",
            sa.Unicode(50),
            sa.ForeignKey("groups.group_name", onupdate="CASCADE", ondelete="SET NULL"),
        ),
        sa.Column(
            "owner_user_name",
            sa.Unicode(30),
            sa.ForeignKey("users.user_name", onupdate="CASCADE", ondelete="SET NULL"),
        ),
    )

    op.create_table(
        "groups_resources_permissions",
        sa.Column(
            "group_name",
            sa.Unicode(50),
            sa.ForeignKey("groups.group_name", onupdate="CASCADE", ondelete="CASCADE"),
            primary_key=True,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger(),
            sa.ForeignKey(
                "resources.resource_id", onupdate="CASCADE", ondelete="CASCADE"
            ),
            primary_key=True,
            autoincrement=False,
        ),
        sa.Column("perm_name", sa.Unicode(50), primary_key=True),
    )

    op.create_table(
        "users_resources_permissions",
        sa.Column(
            "user_name",
            sa.Unicode(50),
            sa.ForeignKey("users.user_name", onupdate="CASCADE", ondelete="CASCADE"),
            primary_key=True,
        ),
        sa.Column(
            "resource_id",
            sa.BigInteger(),
            sa.ForeignKey(
                "resources.resource_id", onupdate="CASCADE", ondelete="CASCADE"
            ),
            primary_key=True,
            autoincrement=False,
        ),
        sa.Column("perm_name", sa.Unicode(50), primary_key=True),
    )


def downgrade():
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declared_attr

from ziggurat_foundations.models.base import BaseModel

__all__ = ["UserGroupMixin"]


class UserGroupMixin(BaseModel):
    """
    Mixin for UserGroup model
    """

    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    @declared_attr
    def __tablename__(self):
        return "users_groups"

    @declared_attr
    def group_id(self):
        return sa.Column(
            sa.Integer,
            sa.ForeignKey("groups.id", onupdate="CASCADE", ondelete="CASCADE"),
            primary_key=True,
        )

    @declared_attr
    def user_id(self):
        return sa.Column(
            sa.Integer,
            sa.ForeignKey("users.id", onupdate="CASCADE", ondelete="CASCADE"),
            primary_key=True,
        )

    def __repr__(self):
        return "<UserGroup: g:%s, u:%s>" % (self.group_id, self.user_id)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import validates

from ziggurat_foundations.models.base import BaseModel

__all__ = ["UserPermissionMixin"]


class UserPermissionMixin(BaseModel):
    """
    Mixin for UserPermission model
    """

    __table_args__ = (
        sa.PrimaryKeyConstraint("user_id", "perm_name", name="pk_users_permissions"),
        {"mysql_engine": "InnoDB", "mysql_charset": "utf8"},
    )

    @declared_attr
    def __tablename__(self):
        return "users_permissions"

    @declared_attr
    def user_id(self):
        return sa.Column(
            sa.Integer,
            sa.ForeignKey("users.id", onupdate="CASCADE", ondelete="CASCADE"),
            primary_key=True,
        )

    @declared_attr
    def perm_name(self):
        return sa.Column(sa.Unicode(64), primary_key=True)

    @validates("perm_name")
    def validate_perm_name(self, key, value):
        if value != value.lower():
            raise AssertionError("perm_name needs to be lowercase")
        return value

    def __repr__(self):
        return "<UserPermission: %s>" % self.perm_name



"""add local_user_id

Revision ID: 57bbf0c387c
Revises: 13391c68750
Create Date: 2016-04-02 09:54:09.643644

"""
from __future__ import unicode_literals

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import table

# revision identifiers, used by Alembic.
revision = "57bbf0c387c"
down_revision = "13391c68750"


def upgrade():
    conn = op.get_bind()
    op.add_column("external_identities", sa.Column("local_user_id", sa.Integer()))

    external_identities_t = table(
        "external_identities",
        sa.Column("local_user_name", sa.Unicode(50)),
        sa.Column("local_user_id", sa.Integer),
    )
    users_t = table(
        "users", sa.Column("user_name", sa.Unicode(50)), sa.Column("id", sa.Integer)
    )

    stmt = (
        external_identities_t.update()
        .values(local_user_id=users_t.c.id)
        .where(users_t.c.user_name == external_identities_t.c.local_user_name)
    )
    conn.execute(stmt)
    op.drop_constraint("pk_external_identities", "external_identities", type="primary")
    op.drop_constraint(
        "fk_external_identities_local_user_name_users",
        "external_identities",
        type="foreignkey",
    )
    op.drop_column("external_identities", "local_user_name")
    op.create_primary_key(
        "pk_external_identities",
        "external_identities",
        cols=["external_id", "local_user_id", "provider_name"],
    )
    op.create_foreign_key(
        None,
        "external_identities",
        "users",
        remote_cols=["id"],
        local_cols=["local_user_id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )


def downgrade():
    pass



"""change primary key sizes

Revision ID: 53927300c277
Revises: 54d08f9adc8c
Create Date: 2012-06-05 23:33:17.943844

"""
from __future__ import unicode_literals

import sqlalchemy as sa
from alembic import op
from alembic.context import get_context
from sqlalchemy.dialects.mysql.base import MySQLDialect
from sqlalchemy.engine.reflection import Inspector

# revision identifiers, used by Alembic.
revision = "53927300c277"
down_revision = "54d08f9adc8c"


def upgrade():
    c = get_context()
    # drop foreign keys for mysql
    if isinstance(c.connection.engine.dialect, MySQLDialect):
        insp = Inspector.from_engine(c.connection.engine)
        for t in [
            "groups_resources_permissions",
            "users_resources_permissions",
            "resources",
        ]:
            for constraint in insp.get_foreign_keys(t):
                if constraint["referred_columns"] == ["resource_id"]:
                    op.drop_constraint(constraint["name"], t, type="foreignkey")

    op.alter_column(
        "resources",
        "resource_id",
        type_=sa.Integer(),
        existing_type=sa.BigInteger(),
        autoincrement=True,
        nullable=False,
    )
    op.alter_column(
        "resources", "parent_id", type_=sa.Integer(), existing_type=sa.BigInteger()
    )
    op.alter_column(
        "users_resources_permissions",
        "resource_id",
        type_=sa.Integer(),
        existing_type=sa.BigInteger(),
        nullable=False,
    )
    op.alter_column(
        "groups_resources_permissions",
        "resource_id",
        type_=sa.Integer(),
        existing_type=sa.BigInteger(),
        nullable=False,
    )

    # recreate foreign keys for mysql
    if isinstance(c.connection.engine.dialect, MySQLDialect):
        op.create_foreign_key(
            "groups_resources_permissions_resource_fk",
            "groups_resources_permissions",
            "resources",
            ["resource_id"],
            ["resource_id"],
            onupdate="CASCADE",
            ondelete="CASCADE",
        )
        op.create_foreign_key(
            "users_resources_permissions_fk",
            "users_resources_permissions",
            "resources",
            ["resource_id"],
            ["resource_id"],
            onupdate="CASCADE",
            ondelete="CASCADE",
        )


def downgrade():
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import os

import pytest
from alembic import command
from alembic.config import Config
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from ziggurat_foundations import ziggurat_model_init
from ziggurat_foundations.models.external_identity import ExternalIdentityMixin
from ziggurat_foundations.models.group import GroupMixin
from ziggurat_foundations.models.group_permission import GroupPermissionMixin
from ziggurat_foundations.models.group_resource_permission import (
    GroupResourcePermissionMixin,
)
from ziggurat_foundations.models.resource import ResourceMixin
from ziggurat_foundations.models.user import UserMixin
from ziggurat_foundations.models.user_group import UserGroupMixin
from ziggurat_foundations.models.user_permission import UserPermissionMixin
from ziggurat_foundations.models.user_resource_permission import (
    UserResourcePermissionMixin,
)
from ziggurat_foundations.permissions import ALL_PERMISSIONS, Allow

not_postgres = "postgres" not in os.environ.get("DB_STRING", "").lower()

Base = declarative_base()


class Group(GroupMixin, Base):
    __possible_permissions__ = (
        "root_administration",
        "administration",
        "backend_admin_panel",
        "manage_apps",
    )


class GroupPermission(GroupPermissionMixin, Base):
    pass


class UserGroup(UserGroupMixin, Base):
    pass


class GroupResourcePermission(GroupResourcePermissionMixin, Base):
    pass


class Resource(ResourceMixin, Base):
    def __acl__(self):
        acls = []

        if self.owner_user_id:
            acls.extend([(Allow, self.owner_user_id, ALL_PERMISSIONS)])

        if self.owner_group_id:
            acls.extend([(Allow, "group:%s" % self.owner_group_id, ALL_PERMISSIONS)])
        return acls


class TestResource(Resource):
    __mapper_args__ = {"polymorphic_identity": "test_resource"}


class TestResourceB(Resource):
    __mapper_args__ = {"polymorphic_identity": "test_resource_b"}


class UserPermission(UserPermissionMixin, Base):
    pass


class UserResourcePermission(UserResourcePermissionMixin, Base):
    pass


class ExternalIdentity(ExternalIdentityMixin, Base):
    pass


class User(UserMixin, Base):
    __possible_permissions__ = ["root", "alter_users", "custom1"]


ziggurat_model_init(
    User,
    Group,
    UserGroup,
    GroupPermission,
    UserPermission,
    UserResourcePermission,
    GroupResourcePermission,
    Resource,
    ExternalIdentity,
)


@pytest.fixture
def db_session(request):
    sql_str = os.environ.get("DB_STRING", "sqlite://")
    engine = create_engine(sql_str)
    engine.echo = True
    # pyramid way
    maker = sessionmaker(bind=engine)
    Base.metadata.bind = engine
    Base.metadata.drop_all(engine)
    engine.execute("DROP TABLE IF EXISTS alembic_ziggurat_foundations_version")
    if sql_str.startswith("sqlite"):
        # sqlite will not work with alembic
        Base.metadata.create_all(engine)
    else:
        alembic_cfg = Config()
        alembic_cfg.set_main_option(
            "script_location", "ziggurat_foundations:migrations"
        )
        alembic_cfg.set_main_option("sqlalchemy.url", sql_str)
        command.upgrade(alembic_cfg, "head")

    session = maker()

    def teardown():
        session.rollback()
        session.close()

    request.addfinalizer(teardown)

    return session


@pytest.fixture
def db_session2(request):
    sql_str = os.environ.get("DB_STRING2", "sqlite://")
    engine = create_engine(sql_str)
    engine.echo = True
    # pyramid way
    maker = sessionmaker(bind=engine)
    Base.metadata.bind = engine
    Base.metadata.drop_all(engine)
    engine.execute("DROP TABLE IF EXISTS alembic_ziggurat_foundations_version")
    if sql_str.startswith("sqlite"):
        # sqlite will not work with alembic
        Base.metadata.create_all(engine)
    else:
        alembic_cfg = Config()
        alembic_cfg.set_main_option(
            "script_location", "ziggurat_foundations:migrations"
        )
        alembic_cfg.set_main_option("sqlalchemy.url", sql_str)
        command.upgrade(alembic_cfg, "head")

    session = maker()

    def teardown():
        session.rollback()
        session.close()

    request.addfinalizer(teardown)

    return session


@pytest.fixture
def pyramid_app(request, db_session):
    from webtest import TestApp
    from pyramid.authentication import AuthTktAuthenticationPolicy
    from pyramid.authorization import ACLAuthorizationPolicy
    from pyramid.config import Configurator
    from ziggurat_foundations.ext.pyramid.sign_in import ZigguratSignInSuccess
    from ziggurat_foundations.ext.pyramid.sign_in import ZigguratSignInBadAuth
    from ziggurat_foundations.ext.pyramid.sign_in import ZigguratSignOut

    auth_policy = AuthTktAuthenticationPolicy("secret")
    authorization_policy = ACLAuthorizationPolicy()

    settings = {"ziggurat_foundations.session_provider_callable": lambda x: db_session}

    config = Configurator(
        settings=settings,
        authentication_policy=auth_policy,
        authorization_policy=authorization_policy,
    )
    config.include("pyramid_jinja2")
    config.include("ziggurat_foundations.ext.pyramid.get_user")
    config.include("ziggurat_foundations.ext.pyramid.sign_in")

    # add ziggurat context handlers
    def sign_in(req):
        user = req.context.user
        for h in req.context.headers:
            req.response.headers.add(h[0], value=h[1])
        return {"view": "sign_in", "username": user.user_name, "user_id": user.id}

    def sign_out(req):
        return {"view": "sign_out"}

    def bad_auth(req):
        return {"view": "bad_auth"}

    def index(req):
        username = None
        if req.user:
            username = req.user.user_name
        return {"view": "index", "username": username}

    config.add_view(sign_in, context=ZigguratSignInSuccess, renderer="json")
    config.add_view(sign_out, context=ZigguratSignOut, renderer="json")
    config.add_view(bad_auth, context=ZigguratSignInBadAuth, renderer="json")
    config.add_route("/", "/")
    config.add_view(index, route_name="/", renderer="json")
    return TestApp(config.make_wsgi_app())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ziggurat_foundations.models.base import get_db_session
from ziggurat_foundations.models.services import BaseService

__all__ = ["UserResourcePermissionService"]


class UserResourcePermissionService(BaseService):
    @classmethod
    def get(cls, user_id, resource_id, perm_name, db_session=None):
        """
        Fetch row using primary key -
        will use existing object in session if already present

        :param user_id:
        :param resource_id:
        :param perm_name:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        return db_session.query(cls.model).get([user_id, resource_id, perm_name])

    @classmethod
    def by_resource_user_and_perm(
        cls, user_id, perm_name, resource_id, db_session=None
    ):
        """
        return all instances by user name, perm name and resource id

        :param user_id:
        :param perm_name:
        :param resource_id:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)

        query = db_session.query(cls.model).filter(cls.model.user_id == user_id)
        query = query.filter(cls.model.resource_id == resource_id)
        query = query.filter(cls.model.perm_name == perm_name)
        return query.first()



"""add security_code_date

Revision ID: 13391c68750
Revises: 438c27ec1c9
Create Date: 2015-06-14 21:28:30.672968

"""
from __future__ import unicode_literals

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "13391c68750"
down_revision = "438c27ec1c9"


def upgrade():
    op.add_column(
        "users",
        sa.Column(
            "security_code_date", sa.DateTime(), server_default="2000-01-01 01:01"
        ),
    )


def downgrade():
    pass



# -*- coding: utf-8 -*-
from __future__ import with_statement, unicode_literals

import six

from ziggurat_foundations.models.services.user import UserService
from ziggurat_foundations.tests import BaseTestCase, add_user
from ziggurat_foundations.tests.conftest import User


class TestUser(BaseTestCase):
    def test_get(self, db_session):
        org_user = add_user(db_session)
        user = UserService.get(user_id=org_user.id, db_session=db_session)
        assert org_user.id == user.id

    def test_add_user(self, db_session):
        user = User(user_name="username", email="email", status=0)
        db_session.add(user)
        db_session.flush()

        user = db_session.query(User).filter(User.user_name == "username")
        user = user.first()
        assert user.user_name == "username"
        assert user.email == "email"
        assert user.status == 0

    def test_delete_user(self, db_session):
        add_user(db_session)
        to_delete = UserService.by_user_name("username", db_session=db_session)
        db_session.delete(to_delete)
        db_session.commit()

    def test_user_repr(self, db_session):
        user = add_user(db_session)
        assert repr(user) == "<User: username>"

    def test_check_password_correct(self, db_session):
        user = add_user(db_session)
        assert UserService.check_password(user, "password") is True

    def test_check_password_wrong(self, db_session):
        user = add_user(db_session)
        assert UserService.check_password(user, "wrong_password") is False

    def test_by_user_name_existing(self, db_session):
        created_user = add_user(db_session)
        queried_user = UserService.by_user_name("username", db_session=db_session)

        assert created_user == queried_user

    def test_by_user_name_not_existing(self, db_session):
        add_user(db_session)
        queried_user = UserService.by_user_name(
            "not_existing_user", db_session=db_session
        )

        assert queried_user is None

    def test_by_user_name_none(self, db_session):
        queried_user = UserService.by_user_name(None, db_session=db_session)

        assert queried_user is None

    def test_by_username_andsecurity_code_existing(self, db_session):
        created_user = add_user(db_session)
        security_code = created_user.security_code
        queried_user = UserService.by_user_name_and_security_code(
            user_name="username", security_code=security_code, db_session=db_session
        )

        assert created_user == queried_user

    def test_by_username_andsecurity_code_not_existing(self, db_session):
        created_user = add_user(db_session)
        security_code = created_user.security_code
        queried_user = UserService.by_user_name_and_security_code(
            user_name="not_existing_user",
            security_code=security_code,
            db_session=db_session,
        )

        assert queried_user is None

    def test_by_username_andsecurity_code_wrong_code(self, db_session):
        add_user(db_session)
        queried_user = UserService.by_user_name_and_security_code(
            user_name="username", security_code="wrong_code", db_session=db_session
        )

        assert queried_user is None

    def test_by_username_andsecurity_code_none(self, db_session):
        created_user = add_user(db_session)
        security_code = created_user.security_code
        found = UserService.by_user_name_and_security_code(
            user_name=None, security_code=security_code, db_session=db_session
        )

        assert found is None

    def test_by_user_names(self, db_session):
        user1 = add_user(db_session, "user1", "email1")
        add_user(db_session, "user2", "email2")
        user3 = add_user(db_session, "user3", "email3")

        queried_users = UserService.by_user_names(
            ["user1", "user3"], db_session=db_session
        ).all()

        assert len(queried_users) == 2
        assert user1 == queried_users[0]
        assert user3 == queried_users[1]

    def test_by_user_names_one_none(self, db_session):
        user1 = add_user(db_session, "user1", "email1")
        add_user(db_session, "user2", "email2")
        user3 = add_user(db_session, "user3", "email3")

        queried_users = UserService.by_user_names(
            ["user1", None, "user3"], db_session=db_session
        ).all()

        assert len(queried_users) == 2
        assert user1 == queried_users[0]
        assert user3 == queried_users[1]

    def test_by_user_names_like(self, db_session):
        user1 = add_user(db_session, "user1", "email1")
        add_user(db_session, "luser2", "email2")
        add_user(db_session, "noname", "email3")

        queried_users = UserService.user_names_like(
            "user%", db_session=db_session
        ).all()
        assert len(queried_users) == 1
        assert user1 == queried_users[0]

    def test_by_user_names_like_none(self, db_session):
        queried_users = UserService.user_names_like(None, db_session=db_session).all()
        assert [] == queried_users

    def test_by_email(self, db_session):
        created_user = add_user(db_session)
        queried_user = UserService.by_email("email", db_session=db_session)

        assert created_user == queried_user

    def test_by_email_none(self, db_session):
        add_user(db_session)
        queried_user = UserService.by_email(None, db_session=db_session)

        assert queried_user is None

    def test_by_email_wrong_email(self, db_session):
        add_user(db_session)
        queried_user = UserService.by_email("wrong_email", db_session=db_session)

        assert queried_user is None

    def test_by_mail_and_username(self, db_session):
        created_user = add_user(db_session)
        queried_user = UserService.by_email_and_username(
            "email", "username", db_session=db_session
        )

        assert created_user == queried_user

    def test_by_mail_and_username_wrong_mail(self, db_session):
        add_user(db_session)
        queried_user = UserService.by_email_and_username(
            "wrong_email", "username", db_session=db_session
        )

        assert queried_user is None

    def test_by_mail_and_username_wrong_username(self, db_session):
        add_user(db_session)
        queried_user = UserService.by_email_and_username(
            "email", "wrong_username", db_session=db_session
        )

        assert queried_user is None

    def test_by_mail_and_username_none(self, db_session):
        found = UserService.by_email_and_username("email", None, db_session=db_session)
        assert found is None

    def test_gravatar_url(self, db_session):
        user = add_user(db_session)
        user.email = "arkadiy@bk.ru"
        assert (
            UserService.gravatar_url(user) == "https://"
            "secure.gravatar.com/avatar/"
            "cbb6777e4a7ec0d96b33d2033e59fec6?d=mm"
        )

    def test_gravatar_url_with_params(self, db_session):
        # pylint: disable=E0401
        import six.moves.urllib.parse as parser

        # pylint: enable=E0401
        user = add_user(db_session)
        user.email = "arkadiy@bk.r"
        gravatar_url = UserService.gravatar_url(user, s=100, r="pg")
        parsed_url = parser.urlparse(gravatar_url)
        qs_dict = parser.parse_qs(parsed_url.query)
        assert qs_dict == {"s": ["100"], "d": ["mm"], "r": ["pg"]}

    def test_generate_random_string(self, db_session):
        rand_str = UserService.generate_random_string()

        assert len(rand_str) == 7
        assert isinstance(rand_str, six.string_types)

    def test_generate_random_pass(self, db_session):
        rand_str = UserService.generate_random_pass()

        assert len(rand_str) == 7
        assert isinstance(rand_str, six.string_types)

        rand_str = UserService.generate_random_pass(20)
        assert len(rand_str) == 20

    def test_regenerate_security_code(self, db_session):
        created_user = add_user(db_session)
        old_code = created_user.security_code
        UserService.regenerate_security_code(created_user)
        new_code = created_user.security_code

        assert old_code != new_code
        assert len(new_code) == 64



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.md")) as f:
    README = f.read()
with open(os.path.join(here, "CHANGES.md")) as f:
    CHANGES = f.read()

test_deps = [
    "coverage",
    "pytest",
    "pytest-cov",
    "tox",
    "mock",
    "pyramid",
    "webtest",
    "pyramid_jinja2",
]

setup(
    name="ziggurat_foundations",
    version="0.8.3",
    description=""" Set of SQLAlchemy mixins that make application building an easy task. Provides users, groups,
    permissions, resource tree handling and authorization solutions for Pyramid and Flask frameworks.
    """,
    long_description_content_type="text/markdown",
    long_description=README,
    author="Marcin Lulek",
    author_email="info@webreactor.eu",
    url="https://github.com/ergo/ziggurat_foundations",
    license="BSD",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Flask",
        "Framework :: Pylons",
        "Framework :: Pyramid",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(),
    zip_safe=True,
    include_package_data=True,
    test_suite="ziggurat_foundations.tests",
    tests_require=test_deps,
    install_requires=[
        "sqlalchemy",
        "passlib>=1.6.1",
        "paginate",
        "paginate_sqlalchemy",
        "alembic",
        "zope.deprecation >= 3.5.0",
        "six",
    ],
    setup_requires=["pytest-runner"],
    extras_require={
        "test": test_deps,
        "lint": ["black", "pylint", "rstcheck", "flake8"],
    },
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sqlalchemy as sa

from ziggurat_foundations.models.base import get_db_session
from ziggurat_foundations.models.services import BaseService
from ziggurat_foundations.permissions import (
    ANY_PERMISSION,
    ALL_PERMISSIONS,
    PermissionTuple,
    resource_permissions_for_users,
)

__all__ = ["ResourceService"]


class ResourceService(BaseService):
    @classmethod
    def get(cls, resource_id, db_session=None):
        """
        Fetch row using primary key -
        will use existing object in session if already present

        :param resource_id:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        return db_session.query(cls.model).get(resource_id)

    @classmethod
    def perms_for_user(cls, instance, user, db_session=None):
        """
        returns all permissions that given user has for this resource
            from groups and directly set ones too

        :param instance:
        :param user:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session, instance)
        query = db_session.query(
            cls.models_proxy.GroupResourcePermission.group_id.label("owner_id"),
            cls.models_proxy.GroupResourcePermission.perm_name,
            sa.literal("group").label("type"),
        )
        query = query.filter(
            cls.models_proxy.GroupResourcePermission.group_id.in_(
                [gr.id for gr in user.groups]
            )
        )
        query = query.filter(
            cls.models_proxy.GroupResourcePermission.resource_id == instance.resource_id
        )

        query2 = db_session.query(
            cls.models_proxy.UserResourcePermission.user_id.label("owner_id"),
            cls.models_proxy.UserResourcePermission.perm_name,
            sa.literal("user").label("type"),
        )
        query2 = query2.filter(
            cls.models_proxy.UserResourcePermission.user_id == user.id
        )
        query2 = query2.filter(
            cls.models_proxy.UserResourcePermission.resource_id == instance.resource_id
        )
        query = query.union(query2)

        groups_dict = dict([(g.id, g) for g in user.groups])
        perms = [
            PermissionTuple(
                user,
                row.perm_name,
                row.type,
                groups_dict.get(row.owner_id) if row.type == "group" else None,
                instance,
                False,
                True,
            )
            for row in query
        ]

        # include all perms if user is the owner of this resource
        if instance.owner_user_id == user.id:
            perms.append(
                PermissionTuple(
                    user, ALL_PERMISSIONS, "user", None, instance, True, True
                )
            )
        groups_dict = dict([(g.id, g) for g in user.groups])
        if instance.owner_group_id in groups_dict:
            perms.append(
                PermissionTuple(
                    user,
                    ALL_PERMISSIONS,
                    "group",
                    groups_dict.get(instance.owner_group_id),
                    instance,
                    True,
                    True,
                )
            )

        return perms

    @classmethod
    def direct_perms_for_user(cls, instance, user, db_session=None):
        """
        returns permissions that given user has for this resource
            without ones inherited from groups that user belongs to

        :param instance:
        :param user:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session, instance)
        query = db_session.query(
            cls.models_proxy.UserResourcePermission.user_id,
            cls.models_proxy.UserResourcePermission.perm_name,
        )
        query = query.filter(cls.models_proxy.UserResourcePermission.user_id == user.id)
        query = query.filter(
            cls.models_proxy.UserResourcePermission.resource_id == instance.resource_id
        )

        perms = [
            PermissionTuple(user, row.perm_name, "user", None, instance, False, True)
            for row in query
        ]

        # include all perms if user is the owner of this resource
        if instance.owner_user_id == user.id:
            perms.append(
                PermissionTuple(user, ALL_PERMISSIONS, "user", None, instance, True)
            )
        return perms

    @classmethod
    def group_perms_for_user(cls, instance, user, db_session=None):
        """
        returns permissions that given user has for this resource
            that are inherited from groups

        :param instance:
        :param user:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session, instance)
        perms = resource_permissions_for_users(
            cls.models_proxy,
            ANY_PERMISSION,
            resource_ids=[instance.resource_id],
            user_ids=[user.id],
            db_session=db_session,
        )
        perms = [p for p in perms if p.type == "group"]
        # include all perms if user is the owner of this resource
        groups_dict = dict([(g.id, g) for g in user.groups])
        if instance.owner_group_id in groups_dict:
            perms.append(
                PermissionTuple(
                    user,
                    ALL_PERMISSIONS,
                    "group",
                    groups_dict.get(instance.owner_group_id),
                    instance,
                    True,
                    True,
                )
            )
        return perms

    @classmethod
    def users_for_perm(
        cls,
        instance,
        perm_name,
        user_ids=None,
        group_ids=None,
        limit_group_permissions=False,
        skip_group_perms=False,
        db_session=None,
    ):
        """
        return PermissionTuples for users AND groups that have given
        permission for the resource, perm_name is __any_permission__ then
        users with any permission will be listed

        :param instance:
        :param perm_name:
        :param user_ids: limits the permissions to specific user ids
        :param group_ids: limits the permissions to specific group ids
        :param limit_group_permissions: should be used if we do not want to have
        user objects returned for group permissions, this might cause performance
        issues for big groups
        :param skip_group_perms: do not attach group permissions to the resultset
        :param db_session:
        :return:
        """  # noqa
        db_session = get_db_session(db_session, instance)
        users_perms = resource_permissions_for_users(
            cls.models_proxy,
            [perm_name],
            [instance.resource_id],
            user_ids=user_ids,
            group_ids=group_ids,
            limit_group_permissions=limit_group_permissions,
            skip_group_perms=skip_group_perms,
            db_session=db_session,
        )
        if instance.owner_user_id:
            users_perms.append(
                PermissionTuple(
                    instance.owner, ALL_PERMISSIONS, "user", None, instance, True, True
                )
            )
        if instance.owner_group_id and not skip_group_perms:
            for user in instance.owner_group.users:
                users_perms.append(
                    PermissionTuple(
                        user,
                        ALL_PERMISSIONS,
                        "group",
                        instance.owner_group,
                        instance,
                        True,
                        True,
                    )
                )

        return users_perms

    @classmethod
    def by_resource_id(cls, resource_id, db_session=None):
        """
        fetch the resouce by id

        :param resource_id:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model).filter(
            cls.model.resource_id == int(resource_id)
        )
        return query.first()

    @classmethod
    def perm_by_group_and_perm_name(
        cls, resource_id, group_id, perm_name, db_session=None
    ):
        """
        fetch permissions by group and permission name

        :param resource_id:
        :param group_id:
        :param perm_name:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.models_proxy.GroupResourcePermission)
        query = query.filter(
            cls.models_proxy.GroupResourcePermission.group_id == group_id
        )
        query = query.filter(
            cls.models_proxy.GroupResourcePermission.perm_name == perm_name
        )
        query = query.filter(
            cls.models_proxy.GroupResourcePermission.resource_id == resource_id
        )
        return query.first()

    @classmethod
    def groups_for_perm(
        cls,
        instance,
        perm_name,
        group_ids=None,
        limit_group_permissions=False,
        db_session=None,
    ):
        """
        return PermissionTuples for groups that have given
        permission for the resource, perm_name is __any_permission__ then
        users with any permission will be listed

        :param instance:
        :param perm_name:
        :param group_ids: limits the permissions to specific group ids
        :param limit_group_permissions: should be used if we do not want to have
        user objects returned for group permissions, this might cause performance
        issues for big groups
        :param db_session:
        :return:
        """  # noqa
        db_session = get_db_session(db_session, instance)
        group_perms = resource_permissions_for_users(
            cls.models_proxy,
            [perm_name],
            [instance.resource_id],
            group_ids=group_ids,
            limit_group_permissions=limit_group_permissions,
            skip_user_perms=True,
            db_session=db_session,
        )
        if instance.owner_group_id:
            for user in instance.owner_group.users:
                group_perms.append(
                    PermissionTuple(
                        user,
                        ALL_PERMISSIONS,
                        "group",
                        instance.owner_group,
                        instance,
                        True,
                        True,
                    )
                )

        return group_perms

    @classmethod
    def lock_resource_for_update(cls, resource_id, db_session):
        """
        Selects resource for update - locking access for other transactions

        :param resource_id:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model)
        query = query.filter(cls.model.resource_id == resource_id)
        query = query.with_for_update()
        return query.first()



# -*- coding: utf-8 -*-
from __future__ import with_statement, unicode_literals

from ziggurat_foundations.tests.conftest import (
    User,
    Group,
    GroupPermission,
    UserPermission,
    UserResourcePermission,
    GroupResourcePermission,
    Resource,
    TestResource,
    TestResourceB,
)
from ziggurat_foundations.models.services.user import UserService


def check_one_in_other(first, second):
    while first:
        item = first.pop()
        ix = second.index(item)
        second.pop(ix)
    assert len(first) == len(second)


def add_user(
    db_session,
    user_name="username",
    email="email",
    perms=["root", "alter_users"],
    password="password",
):
    user = User(user_name=user_name, email=email, status=0)
    UserService.set_password(user, password)
    for perm in perms:
        u_perm = UserPermission(perm_name=perm)
        user.user_permissions.append(u_perm)
    db_session.add(user)
    db_session.flush()
    return user


def add_resource(
    db_session,
    resource_id,
    resource_name="test_resource",
    parent_id=None,
    ordering=None,
):
    Resource.__possible_permissions__ = [
        "test_perm",
        "test_perm1",
        "test_perm2",
        "foo_perm",
        "group_perm",
        "group_perm2",
    ]
    resource = TestResource(
        resource_id=resource_id,
        resource_name=resource_name,
        parent_id=parent_id,
        ordering=ordering,
    )
    db_session.add(resource)
    db_session.flush()
    return resource


def add_resource_b(db_session, resource_id, resource_name="test_resource"):
    Resource.__possible_permissions__ = [
        "test_perm",
        "test_perm1",
        "test_perm2",
        "foo_perm",
        "group_perm",
        "group_perm2",
    ]
    resource = TestResourceB(resource_id=resource_id, resource_name=resource_name)
    db_session.add(resource)
    db_session.flush()
    return resource


def add_group(db_session, group_name="group", description="desc"):
    group = Group(group_name=group_name, description=description)
    test_perm = GroupPermission(perm_name="manage_apps")
    group.permissions.append(test_perm)
    db_session.add(group)
    db_session.flush()
    return group


def create_default_tree(db_session):
    root = add_resource(db_session, -1, "root a", ordering=1)
    res_a = add_resource(db_session, 1, "a", parent_id=root.resource_id, ordering=1)
    add_resource(db_session, 5, "aa", parent_id=res_a.resource_id, ordering=1)
    add_resource(db_session, 6, "ab", parent_id=res_a.resource_id, ordering=2)
    res_ac = add_resource(db_session, 7, "ac", parent_id=res_a.resource_id, ordering=3)
    res_aca = add_resource(
        db_session, 9, "aca", parent_id=res_ac.resource_id, ordering=1
    )
    add_resource(db_session, 12, "acaa", parent_id=res_aca.resource_id, ordering=1)
    add_resource(db_session, 8, "ad", parent_id=res_a.resource_id, ordering=4)
    res_b = add_resource(db_session, 2, "b", parent_id=root.resource_id, ordering=2)
    add_resource(db_session, 4, "ba", parent_id=res_b.resource_id, ordering=1)
    add_resource(db_session, 3, "c", parent_id=root.resource_id, ordering=3)
    add_resource(db_session, 10, "d", parent_id=root.resource_id, ordering=4)
    add_resource(db_session, 11, "e", parent_id=root.resource_id, ordering=5)
    root_b = add_resource(db_session, -2, "root b", ordering=2)
    root_c = add_resource(db_session, -3, "root c", ordering=3)
    return [root, root_b, root_c]


class BaseTestCase(object):
    def set_up_user_group_and_perms(self, db_session):
        """
        perm map:

        username:
            first_user : root, alter_users
            res_perms: r1:g1:foo_perm, r1:g1:test_perm2

        foouser:
            user_perms : custom
            res_perms: r2:foo_perm

        baruser:
            user_perms : root, alter_users
            res_perms: r2:test_perm

        bazuser:
            user_perms : root, alter_users
            res_perms: r1:g2:group_perm

        """
        created_user = add_user(db_session, user_name="first_user")
        created_user2 = add_user(
            db_session, user_name="foouser", email="new_email", perms=["custom"]
        )
        created_user3 = add_user(db_session, user_name="baruser", email="new_email2")
        created_user4 = add_user(db_session, user_name="bazuser", email="new_email3")
        resource = add_resource(db_session, 1, "test_resource")
        resource2 = add_resource_b(db_session, 2, "other_resource")
        group = add_group(db_session)
        group2 = add_group(db_session, group_name="group2")
        group.users.append(created_user)
        group2.users.append(created_user4)
        group_permission = GroupResourcePermission(
            perm_name="group_perm", group_id=group.id
        )
        group_permission2 = GroupResourcePermission(
            perm_name="group_perm", group_id=group2.id
        )
        user_permission = UserResourcePermission(
            perm_name="test_perm2", user_id=created_user.id
        )
        user_permission2 = UserResourcePermission(
            perm_name="foo_perm", user_id=created_user.id
        )
        user2_permission = UserResourcePermission(
            perm_name="foo_perm", user_id=created_user2.id
        )
        user3_permission = UserResourcePermission(
            perm_name="test_perm", user_id=created_user3.id
        )
        resource.group_permissions.append(group_permission)
        resource.group_permissions.append(group_permission2)
        resource.user_permissions.append(user_permission)
        resource.user_permissions.append(user_permission2)
        resource2.user_permissions.append(user2_permission)
        resource2.user_permissions.append(user3_permission)
        db_session.flush()
        self.resource = resource
        self.resource2 = resource2
        self.user = created_user
        self.user2 = created_user2
        self.user3 = created_user3
        self.user4 = created_user4
        self.group = group
        self.group2 = group2


class DummyUserObj(object):
    def __init__(self):
        self.user_name = "new_name"
        self.user_password = "foo"
        self.email = "change@email.com"



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ziggurat_foundations.models.base import get_db_session
from ziggurat_foundations.models.services import BaseService

__all__ = ["ExternalIdentityService"]


class ExternalIdentityService(BaseService):
    @classmethod
    def get(cls, external_id, local_user_id, provider_name, db_session=None):
        """
        Fetch row using primary key -
        will use existing object in session if already present

        :param external_id:
        :param local_user_id:
        :param provider_name:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        return db_session.query(cls.model).get(
            [external_id, local_user_id, provider_name]
        )

    @classmethod
    def by_external_id_and_provider(cls, external_id, provider_name, db_session=None):
        """
        Returns ExternalIdentity instance based on search params

        :param external_id:
        :param provider_name:
        :param db_session:
        :return: ExternalIdentity
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model)
        query = query.filter(cls.model.external_id == external_id)
        query = query.filter(cls.model.provider_name == provider_name)
        return query.first()

    @classmethod
    def user_by_external_id_and_provider(
        cls, external_id, provider_name, db_session=None
    ):
        """
        Returns User instance based on search params

        :param external_id:
        :param provider_name:
        :param db_session:
        :return: User
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.models_proxy.User)
        query = query.filter(cls.model.external_id == external_id)
        query = query.filter(cls.model.provider_name == provider_name)
        query = query.filter(cls.models_proxy.User.id == cls.model.local_user_id)
        return query.first()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import importlib
import logging

from ziggurat_foundations.models.base import get_db_session
from ziggurat_foundations.models.services.user import UserService

CONFIG_KEY = "ziggurat_foundations"
log = logging.getLogger(__name__)


def includeme(config):
    settings = config.registry.settings

    session_provider_callable_config = settings.get(
        "%s.session_provider_callable" % CONFIG_KEY
    )
    try_global_session = False

    if not session_provider_callable_config:

        def session_provider_callable(request):
            return get_db_session()

        try_global_session = True
    else:
        if callable(session_provider_callable_config):
            session_provider_callable = session_provider_callable_config
        else:
            parts = session_provider_callable_config.split(":")
            _tmp = importlib.import_module(parts[0])
            session_provider_callable = getattr(_tmp, parts[1])

    # This function is bundled into the request, so for each request you can
    # do request.user
    def get_user(request):
        userid = request.unauthenticated_userid
        if try_global_session:
            # set db_session to none to pass to the UserModel.by_id so it can try to autodiscover
            db_session = None
        else:
            # Else assign the request.session
            db_session = session_provider_callable(request)
        if userid is not None:
            return UserService.by_id(userid, db_session=db_session)

    # add in request.user function
    config.add_request_method(get_user, "user", reify=True, property=True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declared_attr

from ziggurat_foundations.models.base import BaseModel


__all__ = ["GroupMixin"]


class GroupMixin(BaseModel):
    """ Mixin for Group model """

    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    @declared_attr
    def __tablename__(self):
        return "groups"

    # lists app wide permissions we might want to assign to groups
    __possible_permissions__ = ()

    @declared_attr
    def id(self):
        return sa.Column(sa.Integer(), primary_key=True)

    @declared_attr
    def group_name(self):
        return sa.Column(sa.Unicode(128), nullable=False, unique=True)

    @declared_attr
    def description(self):
        return sa.Column(sa.Text())

    @declared_attr
    def member_count(self):
        return sa.Column(sa.Integer, nullable=False, default=0)

    @declared_attr
    def users(self):
        """ relationship for users belonging to this group"""
        return sa.orm.relationship(
            "User",
            secondary="users_groups",
            order_by="User.user_name",
            passive_deletes=True,
            passive_updates=True,
            backref="groups",
        )

    # dynamic property - useful
    @declared_attr
    def users_dynamic(self):
        """ dynamic relationship for users belonging to this group
            one can use filter """
        return sa.orm.relationship(
            "User", secondary="users_groups", order_by="User.user_name", lazy="dynamic"
        )

    @declared_attr
    def permissions(self):
        """ non-resource permissions assigned to this group"""
        return sa.orm.relationship(
            "GroupPermission",
            backref="groups",
            cascade="all, delete-orphan",
            passive_deletes=True,
            passive_updates=True,
        )

    @declared_attr
    def resource_permissions(self):
        """ permissions to specific resources this group has"""
        return sa.orm.relationship(
            "GroupResourcePermission",
            backref="groups",
            cascade="all, delete-orphan",
            passive_deletes=True,
            passive_updates=True,
        )

    @declared_attr
    def resources(self):
        """ Returns all resources directly owned by group, can be used to assign
        ownership of new resources::

            user.resources.append(resource) """
        return sa.orm.relationship(
            "Resource",
            cascade="all",
            passive_deletes=True,
            passive_updates=True,
            backref="owner_group",
        )

    @declared_attr
    def resources_dynamic(self):
        """ Returns all resources directly owned by group, can be used to assign
        ownership of new resources::

            user.resources.append(resource) """
        return sa.orm.relationship(
            "Resource",
            cascade="all",
            passive_deletes=True,
            passive_updates=True,
            lazy="dynamic",
        )

    @sa.orm.validates("permissions")
    def validate_permission(self, key, permission):
        """ validates if group can get assigned with permission"""
        if permission.perm_name not in self.__possible_permissions__:
            raise AssertionError(
                "perm_name is not one of {}".format(self.__possible_permissions__)
            )
        return permission

    def __repr__(self):
        return "<Group: %s, %s>" % (self.group_name, self.id)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ziggurat_foundations.models.base import get_db_session
from ziggurat_foundations.models.services import BaseService

__all__ = ["GroupResourcePermissionService"]


class GroupResourcePermissionService(BaseService):
    @classmethod
    def get(cls, group_id, resource_id, perm_name, db_session=None):
        """
        Fetch row using primary key -
        will use existing object in session if already present

        :param group_id:
        :param resource_id:
        :param perm_name:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        return db_session.query(cls.model).get([group_id, resource_id, perm_name])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import hashlib

import six
import sqlalchemy as sa

from ziggurat_foundations.models.base import get_db_session
from ziggurat_foundations.models.services import BaseService
from ziggurat_foundations.permissions import (
    ALL_PERMISSIONS,
    ANY_PERMISSION,
    PermissionTuple,
    resource_permissions_for_users,
)
from ziggurat_foundations.utils import generate_random_string

__all__ = ["UserService"]


class UserService(BaseService):
    @classmethod
    def get(cls, user_id, db_session=None):
        """
        Fetch row using primary key -
        will use existing object in session if already present

        :param user_id:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        return db_session.query(cls.model).get(user_id)

    @classmethod
    def permissions(cls, instance, db_session=None):
        """
        returns all non-resource permissions based on what groups user
            belongs and directly set ones for this user

        :param instance:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session, instance)
        query = db_session.query(
            cls.models_proxy.GroupPermission.group_id.label("owner_id"),
            cls.models_proxy.GroupPermission.perm_name.label("perm_name"),
            sa.literal("group").label("type"),
        )
        query = query.filter(
            cls.models_proxy.GroupPermission.group_id
            == cls.models_proxy.UserGroup.group_id
        )
        query = query.filter(
            cls.models_proxy.User.id == cls.models_proxy.UserGroup.user_id
        )
        query = query.filter(cls.models_proxy.User.id == instance.id)

        query2 = db_session.query(
            cls.models_proxy.UserPermission.user_id.label("owner_id"),
            cls.models_proxy.UserPermission.perm_name.label("perm_name"),
            sa.literal("user").label("type"),
        )
        query2 = query2.filter(cls.models_proxy.UserPermission.user_id == instance.id)
        query = query.union(query2)
        groups_dict = dict([(g.id, g) for g in instance.groups])
        return [
            PermissionTuple(
                instance,
                row.perm_name,
                row.type,
                groups_dict.get(row.owner_id) if row.type == "group" else None,
                None,
                False,
                True,
            )
            for row in query
        ]

    @classmethod
    def resources_with_perms(
        cls, instance, perms, resource_ids=None, resource_types=None, db_session=None
    ):
        """
        returns all resources that user has perms for
            (note that at least one perm needs to be met)

        :param instance:
        :param perms:
        :param resource_ids: restricts the search to specific resources
        :param resource_types:
        :param db_session:
        :return:
        """
        # owned entities have ALL permissions so we return those resources too
        # even without explict perms set
        # TODO: implement admin superrule perm - maybe return all apps
        db_session = get_db_session(db_session, instance)
        query = db_session.query(cls.models_proxy.Resource).distinct()
        group_ids = [gr.id for gr in instance.groups]
        # if user has some groups lets try to join based on their permissions
        if group_ids:
            join_conditions = (
                cls.models_proxy.GroupResourcePermission.group_id.in_(group_ids),
                cls.models_proxy.Resource.resource_id
                == cls.models_proxy.GroupResourcePermission.resource_id,
                cls.models_proxy.GroupResourcePermission.perm_name.in_(perms),
            )
            query = query.outerjoin(
                (cls.models_proxy.GroupResourcePermission, sa.and_(*join_conditions))
            )
            # ensure outerjoin permissions are correct -
            # dont add empty rows from join
            # conditions are - join ON possible group permissions
            # OR owning group/user
            query = query.filter(
                sa.or_(
                    cls.models_proxy.Resource.owner_user_id == instance.id,
                    cls.models_proxy.Resource.owner_group_id.in_(group_ids),
                    cls.models_proxy.GroupResourcePermission.perm_name != None,
                )  # noqa
            )
        else:
            # filter just by username
            query = query.filter(cls.models_proxy.Resource.owner_user_id == instance.id)
        # lets try by custom user permissions for resource
        query2 = db_session.query(cls.models_proxy.Resource).distinct()
        query2 = query2.filter(
            cls.models_proxy.UserResourcePermission.user_id == instance.id
        )
        query2 = query2.filter(
            cls.models_proxy.Resource.resource_id
            == cls.models_proxy.UserResourcePermission.resource_id
        )
        query2 = query2.filter(
            cls.models_proxy.UserResourcePermission.perm_name.in_(perms)
        )
        if resource_ids:
            query = query.filter(
                cls.models_proxy.Resource.resource_id.in_(resource_ids)
            )
            query2 = query2.filter(
                cls.models_proxy.Resource.resource_id.in_(resource_ids)
            )

        if resource_types:
            query = query.filter(
                cls.models_proxy.Resource.resource_type.in_(resource_types)
            )
            query2 = query2.filter(
                cls.models_proxy.Resource.resource_type.in_(resource_types)
            )
        query = query.union(query2)
        query = query.order_by(cls.models_proxy.Resource.resource_name)
        return query

    @classmethod
    def groups_with_resources(cls, instance):
        """
        Returns a list of groups users belongs to with eager loaded
        resources owned by those groups

        :param instance:
        :return:
        """
        return instance.groups_dynamic.options(
            sa.orm.eagerload(cls.models_proxy.Group.resources)
        )

    @classmethod
    def resources_with_possible_perms(
        cls, instance, resource_ids=None, resource_types=None, db_session=None
    ):
        """
        returns list of permissions and resources for this user

        :param instance:
        :param resource_ids: restricts the search to specific resources
        :param resource_types: restricts the search to specific resource types
        :param db_session:
        :return:
        """
        perms = resource_permissions_for_users(
            cls.models_proxy,
            ANY_PERMISSION,
            resource_ids=resource_ids,
            resource_types=resource_types,
            user_ids=[instance.id],
            db_session=db_session,
        )
        for resource in instance.resources:
            perms.append(
                PermissionTuple(
                    instance, ALL_PERMISSIONS, "user", None, resource, True, True
                )
            )
        for group in cls.groups_with_resources(instance):
            for resource in group.resources:
                perms.append(
                    PermissionTuple(
                        instance, ALL_PERMISSIONS, "group", group, resource, True, True
                    )
                )

        return perms

    @classmethod
    def gravatar_url(cls, instance, default="mm", **kwargs):
        """
        returns user gravatar url

        :param instance:
        :param default:
        :param kwargs:
        :return:
        """
        # construct the url
        hash = hashlib.md5(instance.email.encode("utf8").lower()).hexdigest()
        if "d" not in kwargs:
            kwargs["d"] = default
        params = "&".join(
            [
                six.moves.urllib.parse.urlencode({key: value})
                for key, value in kwargs.items()
            ]
        )
        return "https://secure.gravatar.com/avatar/{}?{}".format(hash, params)

    @classmethod
    def set_password(cls, instance, raw_password):
        """
        sets new password on a user using password manager

        :param instance:
        :param raw_password:
        :return:
        """
        # support API for both passlib 1.x and 2.x
        hash_callable = getattr(
            instance.passwordmanager, "hash", instance.passwordmanager.encrypt
        )
        password = hash_callable(raw_password)
        if six.PY2:
            instance.user_password = password.decode("utf8")
        else:
            instance.user_password = password
        cls.regenerate_security_code(instance)

    @classmethod
    def check_password(cls, instance, raw_password, enable_hash_migration=True):
        """
        checks string with users password hash using password manager

        :param instance:
        :param raw_password:
        :param enable_hash_migration: if legacy hashes should be migrated
        :return:
        """
        verified, replacement_hash = instance.passwordmanager.verify_and_update(
            raw_password, instance.user_password
        )
        if enable_hash_migration and replacement_hash:
            if six.PY2:
                instance.user_password = replacement_hash.decode("utf8")
            else:
                instance.user_password = replacement_hash
        return verified

    @classmethod
    def generate_random_pass(cls, chars=7):
        """
        generates random string of fixed length

        :param chars:
        :return:
        """
        return cls.generate_random_string(chars)

    @classmethod
    def regenerate_security_code(cls, instance):
        """
        generates new security code

        :param instance:
        :return:
        """
        instance.security_code = cls.generate_random_string(64)

    @staticmethod
    def generate_random_string(chars=7):
        """

        :param chars:
        :return:
        """
        return generate_random_string(chars)

    @classmethod
    def by_id(cls, user_id, db_session=None):
        """
        fetch user by user id

        :param user_id:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model)
        query = query.filter(cls.model.id == user_id)
        query = query.options(sa.orm.eagerload("groups"))
        return query.first()

    @classmethod
    def by_user_name(cls, user_name, db_session=None):
        """
        fetch user by user name

        :param user_name:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model)
        query = query.filter(
            sa.func.lower(cls.model.user_name) == (user_name or "").lower()
        )
        query = query.options(sa.orm.eagerload("groups"))
        return query.first()

    @classmethod
    def by_user_name_and_security_code(cls, user_name, security_code, db_session=None):
        """
        fetch user objects by user name and security code

        :param user_name:
        :param security_code:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model)
        query = query.filter(
            sa.func.lower(cls.model.user_name) == (user_name or "").lower()
        )
        query = query.filter(cls.model.security_code == security_code)
        return query.first()

    @classmethod
    def by_user_names(cls, user_names, db_session=None):
        """
        fetch user objects by user names

        :param user_names:
        :param db_session:
        :return:
        """
        user_names = [(name or "").lower() for name in user_names]
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model)
        query = query.filter(sa.func.lower(cls.model.user_name).in_(user_names))
        # q = q.options(sa.orm.eagerload(cls.groups))
        return query

    @classmethod
    def user_names_like(cls, user_name, db_session=None):
        """
        fetch users with similar names using LIKE clause

        :param user_name:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model)
        query = query.filter(
            sa.func.lower(cls.model.user_name).like((user_name or "").lower())
        )
        query = query.order_by(cls.model.user_name)
        # q = q.options(sa.orm.eagerload('groups'))
        return query

    @classmethod
    def by_email(cls, email, db_session=None):
        """
        fetch user object by email

        :param email:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model).filter(
            sa.func.lower(cls.model.email) == (email or "").lower()
        )
        query = query.options(sa.orm.eagerload("groups"))
        return query.first()

    @classmethod
    def by_email_and_username(cls, email, user_name, db_session=None):
        """
        fetch user object by email and username

        :param email:
        :param user_name:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model).filter(cls.model.email == email)
        query = query.filter(
            sa.func.lower(cls.model.user_name) == (user_name or "").lower()
        )
        query = query.options(sa.orm.eagerload("groups"))
        return query.first()

    @classmethod
    def users_for_perms(cls, perm_names, db_session=None):
        """
        return users hat have one of given permissions

        :param perm_names:
        :param db_session:
        :return:
        """
        db_session = get_db_session(db_session)
        query = db_session.query(cls.model)
        query = query.filter(
            cls.models_proxy.User.id == cls.models_proxy.UserGroup.user_id
        )
        query = query.filter(
            cls.models_proxy.UserGroup.group_id
            == cls.models_proxy.GroupPermission.group_id
        )
        query = query.filter(cls.models_proxy.GroupPermission.perm_name.in_(perm_names))

        query2 = db_session.query(cls.model)
        query2 = query2.filter(
            cls.models_proxy.User.id == cls.models_proxy.UserPermission.user_id
        )
        query2 = query2.filter(
            cls.models_proxy.UserPermission.perm_name.in_(perm_names)
        )
        users = query.union(query2).order_by(cls.model.id)
        return users
